/*
 * Copyright (c) 1999, 2001, Oracle and/or its affiliates. All rights reserved.
 * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */

package java.lang;

import sun.misc.Signal;
import sun.misc.SignalHandler;


/**
 * Package-private utility class for setting up and tearing down
 * platform-specific support for termination-triggered shutdowns.
 *
 * @author   Mark Reinhold
 * @since    1.3
 */

class Terminator {

    private static SignalHandler handler = null;

    /* Invocations of setup and teardown are already synchronized
     * on the shutdown lock, so no further synchronization is needed here
     */

    static void setup() {
        if (handler != null) return;
        SignalHandler sh = new SignalHandler() {
            public void handle(Signal sig) {
                Shutdown.exit(sig.getNumber() + 0200);
            }
        };
        handler = sh;
	// ibm@144598 starts                                                    //IBM-T6_sun_bug_fixes_temp
        try {
     	    Signal.handle(new Signal("HUP"), sh);                          //IBM-T6_sun_bug_fixes_temp
	} catch (IllegalArgumentException e) {                                  //IBM-T6_sun_bug_fixes_temp
            // When -Xrs is specified the user is responsible for               //IBM-T6_sun_bug_fixes_temp
            // ensuring that shutdown hooks are run by calling                  //IBM-T6_sun_bug_fixes_temp
            // System.exit()                                                    //IBM-T6_sun_bug_fixes_temp
        }                                                                       //IBM-T6_sun_bug_fixes_temp
                                                                                //IBM-T6_sun_bug_fixes_temp
	try {                                                                   //IBM-T6_sun_bug_fixes_temp
	    Signal.handle(new Signal("INT"), sh);                               //IBM-T6_sun_bug_fixes_temp
	} catch (IllegalArgumentException e) {                                  //IBM-T6_sun_bug_fixes_temp
            // When -Xrs is specified the user is responsible for               //IBM-T6_sun_bug_fixes_temp
            // ensuring that shutdown hooks are run by calling                  //IBM-T6_sun_bug_fixes_temp
            // System.exit()                                                    //IBM-T6_sun_bug_fixes_temp
        }                                                                       //IBM-T6_sun_bug_fixes_temp
                                                                                //IBM-T6_sun_bug_fixes_temp
	try {                                                                   //IBM-T6_sun_bug_fixes_temp
                                                                                //IBM-T6_sun_bug_fixes_temp
	Signal.handle(new Signal("TERM"), sh);                                  //IBM-T6_sun_bug_fixes_temp
        } catch (IllegalArgumentException e) {
            // When -Xrs is specified the user is responsible for
            // ensuring that shutdown hooks are run by calling
            // System.exit()
        }
	// ibm@144598 ends                                                      //IBM-T6_sun_bug_fixes_temp
    }

    static void teardown() {
        /* The current sun.misc.Signal class does not support
         * the cancellation of handlers
         */
    }

}
//IBM-T6_sun_bug_fixes_temp
