/*
 * Copyright (c) 1995, 2011, Oracle and/or its affiliates. All rights reserved.
 * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */

/**
 * =======================================================================
 * Module Information:
 *
 * DESCRIPTION: This interface is used to define new networks that will be
 * supported for java.net.Socket.
 * 
 * This package private interface provides RDMA_CAPABILITY.
 * =======================================================================
 */

package java.net;

import java.io.IOException;

interface NetworkProvider {
	/**
	 * Initialize the underlying network.
	 * This method should perform all the steps required to initialize
	 * the components specific to this network.
	 * Steps may include -
	 *   loading the underlying system libraries
	 *   checking whether the network interface hardware is enabled
	 * Successful completion of this method indicates that
	 * the network is ready to be used.
	 * 
	 * @throws IOException to indicate failure during initialization.
	 */
	void initialize() throws IOException;
	
	/**
	 * This method frees resources that are allocated during
	 * the network initialization.
	 */
	void cleanup();
	
	/**
	 * Create a socket impl of the type supported by a specific network
	 * provider implementing this interface.
	 */
	SocketImpl createImpl(SocketImpl oldImpl);
	
	/**
	 * Set flags/fields after an accept() call.
	 * 
	 * @throws IOException to indicate failure in setting flags/fields.
	 */
	void postAccept(SocketImpl serverSocket, SocketImpl socket) throws IOException;

        /**
         * Get the name of the provider
         */
        String getName();
}
