/*
 * Copyright (c) 1995, 2011, Oracle and/or its affiliates. All rights reserved.
 * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */

/**
 * =======================================================================
 * Module Information:
 *
 * DESCRIPTION: This class defines the RDMA network.
 * 
 * This package private class provides RDMA_CAPABILITY.
 * =======================================================================
 */

package java.net;

import java.io.IOException;
import java.security.AccessController;
import sun.security.action.GetPropertyAction;

class RDMANetworkProvider implements NetworkProvider {
	// native methods
	private native int initialize0();
	private native int cleanup0();
	// this provider name
	private static final String name = "RDMA";
	
	static {
		java.security.AccessController.doPrivileged(
				new sun.security.action.LoadLibraryAction("jsor"));
	}

	
	/**
	 * initialize the RDMA network
	 */
	public void initialize() throws IOException {
		if (initialize0() != 0) {
			throw new IOException("Could not initialize RDMA transport!");
		}
	}
	
	/**
	 * destroy RDMA handles that are assigned during
	 * network initialization
	 */
	public void cleanup() {
		cleanup0();
	}
	
	/**
	 * create an RDMA socket impl
	 */
	public SocketImpl createImpl(SocketImpl oldImpl) {
		return new RDMASocketImpl(oldImpl);
	}
	
	/**
	 * do post-accept operations for an RDMA connection
	 */
	public void postAccept(SocketImpl serverSocketImpl, SocketImpl acceptedSocketImpl) throws IOException {
		long serverSocketDataPtr = ((RDMASocketImpl)serverSocketImpl).getSocketDataPtr();
		long clientSocketDataPtr = ((RDMASocketImpl)serverSocketImpl).socketPostAccept(acceptedSocketImpl, serverSocketDataPtr);
		((RDMASocketImpl)acceptedSocketImpl).setSocketDataPtr(clientSocketDataPtr);
	}
	
	/**
	 * get the name of this network provider 
	 */
	public String getName() {
		return name;
	}
}
